import numpy as np
import cuqi
import cuqipy_cil
from cil.framework import ImageGeometry, AcquisitionGeometry, DataContainer
import subprocess # To check for CUDA

# Try importing astra projector
try:
    from cil.plugins.astra import ProjectionOperator as ProjectionOperatorAstra
except ImportError:
    ProjectionOperatorAstra = None

# Try importing tigre projector
try:
    from cil.plugins.tigre import ProjectionOperator as ProjectionOperatorTigre
except ImportError:
    ProjectionOperatorTigre = None

class CILModel(cuqi.model.LinearModel):
    """ Base class of cuqi model using CIL for CT projectors.

    Parameters
    -----------
    acquisition_geometry : CIL acquisition geometry.
        See CIL documentation.

    image_geometry : CIL image geometry.
        See CIL documentation.

    Attributes
    -----------
    range_geometry : cuqi.geometry.Image2D
        The geometry representing the range associated with sinogram.

    domain_geometry : cuqi.geometry.Image2D
        The geometry representing the domain associated with input image.

    ProjectionOperator : CIL ProjectionOperator
        The projection operator handling forwad and adjoint operations.

    Methods
    -----------
    :meth:`forward` the forward operator.
    :meth:`adjoint` the adjoint operator.

    """
    
    def __init__(self, acquisition_geometry: AcquisitionGeometry, image_geometry: ImageGeometry) -> None:

        # Define image geometries
        range_geometry = cuqi.geometry.Image2D(acquisition_geometry.shape)
        domain_geometry = cuqi.geometry.Image2D(image_geometry.shape)
        super().__init__(self._forward_func, self._adjoint_func, domain_geometry=domain_geometry, range_geometry=range_geometry)

        # Create projection operator depending on the backend
        if cuqipy_cil.config.PROJECTION_BACKEND == "astra":

            # If None module is not available
            if ProjectionOperatorAstra is None:
                raise ImportError("Unable to load astra package needed by cil projector! Did you install cil with astra support?")

            self._ProjectionOperator = ProjectionOperatorAstra(image_geometry, acquisition_geometry, device=cuqipy_cil.config.PROJECTION_BACKEND_DEVICE)
        
        elif cuqipy_cil.config.PROJECTION_BACKEND == "tigre":
            
            # If None module is not available
            if ProjectionOperatorTigre is None:
                raise ImportError("Unable to load tigre package needed by cil projector! Did you install cil with tigre support?")

            # If CPU is requested throw error
            if cuqipy_cil.config.PROJECTION_BACKEND_DEVICE == "cpu":
                raise NotImplementedError("Tigre CPU projector not implemented yet! Try using astra backend instead.")

            self._ProjectionOperator = ProjectionOperatorTigre(image_geometry, acquisition_geometry) # no device option for tigre
        
        # Allocate data containers for efficiency
        self._acquisition_data = acquisition_geometry.allocate()
        self._image_data = image_geometry.allocate()

    
    @property
    def acquisition_geometry(self):
        """ The CIL acquisition geometry. """
        return self.ProjectionOperator.range_geometry()

    @property
    def image_geometry(self):
        """ The CIL image geometry. """
        return self.ProjectionOperator.domain_geometry()

    @property
    def ProjectionOperator(self):
        """ The CIL projection operator. """
        return self._ProjectionOperator

    def _forward_func(self, x: np.ndarray) -> np.ndarray:
        self._fill_container_from_numpy(x, self._image_data)
        self.ProjectionOperator.direct(self._image_data, out=self._acquisition_data)
        return self._acquisition_data.as_array()

    def _adjoint_func(self, x: np.ndarray) -> np.ndarray:
        self._fill_container_from_numpy(x, self._acquisition_data)
        self.ProjectionOperator.adjoint(self._acquisition_data, out=self._image_data)
        return self._image_data.as_array()

    @staticmethod
    def _fill_container_from_numpy(array: np.ndarray, container: DataContainer):
        """ Fill a numpy array into a CIL data container without creating a copy. """

        # Check shape
        if array.shape != container.shape:
            raise ValueError("Array shape does not match container shape.")
            
        # Convert to dtype of container only if necessary (this is the main cost)
        if array.dtype != container.dtype: array = array.astype(container.dtype) 

        # Storing directly in .array avoids copying
        container.array = array

class ParallelBeam2DModel(CILModel):
    """ 2D CT model with parallel beam.

    Parameters
    ----------    
    im_size : tuple of ints
        Dimensions of image in pixels.
    
    det_count : int
        Number of detector elements.
       
    angles : ndarray
        Angles of projections, in radians.

    det_spacing : float, default 1
        Detector element size/spacing.

    domain : tuple, default im_size
        Size of image domain.

    """
    
    def __init__(self,
        im_size = (45,45),
        det_count = 50,
        angles = np.linspace(0,np.pi,60),
        det_spacing = None,
        domain = None
        ):

        if domain == None:
            domain = im_size

        if det_spacing is None:
            det_spacing = 1
            # TODO. Use default det_spacing that ensures whole domain is covered.
            # This allows changing im_size or det_count while keeping the same default scan area.
            # Current implementation with tigre causes some severe artifacts if we modify det_spacing.
            # det_spacing = np.sqrt(2) * np.max(domain) / det_count

        # Setup cil geometries for parallel beam CT
        acquisition_geometry = AcquisitionGeometry.create_Parallel2D()
        acquisition_geometry.set_angles(angles, angle_unit="radian")
        acquisition_geometry.set_panel(det_count, pixel_size=det_spacing)
        
        # Setup image geometry
        image_geometry = ImageGeometry(
            voxel_num_x=im_size[0],
            voxel_num_y=im_size[1],
            voxel_size_x=domain[0] / im_size[0],
            voxel_size_y=domain[1] / im_size[1],
        )

        super().__init__(acquisition_geometry, image_geometry)

class FanBeam2DModel(CILModel):
    """ 2D CT model with fan beam.
    
    Assumes a centered beam.

    Parameters
    ------------    
    im_size : tuple of ints
        Dimensions of image in pixels.
    
    det_count : int
        Number of detector elements.
    
    angles : ndarray
        Angles of projections, in radians.

    source_object_dist : scalar
        Distance between source and object.

    object_detector_dist : scalar
        Distance between detector and object.

    det_spacing : int, default 1
        Detector element size/spacing.

    domain : tuple, default im_size
        Size of image domain.

    """
    
    def __init__(self,
        im_size = (45,45),
        det_count = 50,
        angles = np.linspace(0,2*np.pi,60),
        source_object_dist = 200,
        object_detector_dist = 30,
        det_spacing = None,
        domain = None
        ):

        if domain == None:
            domain = im_size

        if det_spacing is None:
            det_spacing = 1

        # Setup cil geometries for parallel beam CT 
        acquisition_geometry = AcquisitionGeometry.create_Cone2D(
            source_position=[0.0, -source_object_dist],
            detector_position=[0.0, object_detector_dist],
        )
        acquisition_geometry.set_angles(angles, angle_unit="radian")
        acquisition_geometry.set_panel(det_count, pixel_size=det_spacing)

        # Setup image geometry
        image_geometry = ImageGeometry(
            voxel_num_x=im_size[0],
            voxel_num_y=im_size[1],
            voxel_size_x=domain[0] / im_size[0],
            voxel_size_y=domain[1] / im_size[1],
        )

        super().__init__(acquisition_geometry, image_geometry)

class ShiftedFanBeam2DModel(CILModel):
    """
    2D CT model with fanbeam and source+detector shift, assuming object is at position (0,0)

    Parameters
    ------------    
    im_size : tuple of ints
        Dimensions of image in pixels.
    
    det_count : int
        Number of detector elements.
       
    angles : ndarray
        Angles of projections, in radians.

    source_y : scalar
        Source position on y-axis.

    detector_y : scalar
        Detector position on y-axis.

    beamshift_x : scalar
        Source and detector position on x-axis.

    det_spacing : int, default 1
        Detector element size/spacing.

    domain : tuple
        Size of image domain, default (550,550).

    """

    def __init__(self,
        im_size = (45,45),
        det_count = 50,
        angles = np.linspace(0,2*np.pi,60),
        source_y = -600,
        detector_y = 500,
        beamshift_x = -125.3,
        domain = (550,550),
        det_spacing = None
        ):

        if det_spacing is None:
            det_spacing = 1

        # Setup cil geometries for parallel beam CT with shifted source and detector
        acquisition_geometry = AcquisitionGeometry.create_Cone2D(
            source_position=[beamshift_x, source_y],
            detector_position=[beamshift_x, detector_y]
        )
        acquisition_geometry.set_angles(angles, angle_unit="radian")
        acquisition_geometry.set_panel(det_count, pixel_size=det_spacing)

        # Setup image geometry
        image_geometry = ImageGeometry(
            voxel_num_x=im_size[0],
            voxel_num_y=im_size[1],
            voxel_size_x=domain[0] / im_size[0],
            voxel_size_y=domain[1] / im_size[1],
        )

        super().__init__(acquisition_geometry, image_geometry)



import cuqipy_cil
import pytest
import cuqi
import numpy as np
import matplotlib.pyplot as plt

def test_testproblem_simple():
    # Create simple testproblem
    TP = cuqipy_cil.testproblem.ParallelBeam2D()

    # First check dimension of model matches dimension of image + data
    assert TP.model.domain_dim == TP.exactSolution.geometry.par_dim

    # Check that the output is the correct shape
    assert TP.model.range_dim == TP.exactData.geometry.par_dim

    # Check basic run posterior sampling
    samples = TP.UQ(Ns=20)

    # Check that the output is the correct shape
    assert samples.shape == (TP.model.domain_dim, 20)

    # Check basic MAP estimate
    MAP = TP.MAP()

    # Check that the output is the correct shape
    assert MAP.shape == (TP.model.domain_dim,)

@pytest.mark.parametrize("phantom",
    [
        (cuqi.data.shepp_logan(size=128)),
        (cuqi.data.p_power(size=128)),
        (cuqi.data.grains(size=128)),
        ("shepp_logan"),
        ("p_power"),
        ("grains"),
    ])
def test_testproblem_phantom(phantom):
    # Create simple testproblem
    TP = cuqipy_cil.testproblem.ParallelBeam2D(im_size=(128,128), phantom=phantom)

    if isinstance(phantom, str): # If phantom is a string, it should be a valid phantom name
        phantom = cuqi.data.__dict__[phantom](size=TP.model.domain_geometry.fun_shape[0])

    assert np.allclose(TP.exactSolution, cuqi.data.imresize(phantom, TP.model.domain_geometry.fun_shape))

def test_testproblem_set_prior():
    """ Test if one can set a prior after creating a testproblem """
    TP = cuqipy_cil.testproblem.ParallelBeam2D(
        im_size=(256, 256),
        det_count=256,
        angles=np.linspace(0, np.pi, 180),
        phantom="shepp-logan",
    )

    # Cauchy difference prior
    TP.prior = cuqi.distribution.CMRF(
        location=0,
        scale=0.01,
        geometry=TP.model.domain_geometry,
    )

    # Sample posterior with automatic sampler choice
    samples = TP.sample_posterior(20)

    # Plot sample mean and ci
    samples.plot_ci()

def test_testproblem_from_readme():

    # Load a CT forward model and data from testproblem library
    A, y_data, info = cuqipy_cil.testproblem.ParallelBeam2D(
        im_size=(128, 128),
        det_count=128,
        angles=np.linspace(0, np.pi, 180),
        phantom="shepp-logan"
    ).get_components()

    # Set up Bayesian model
    x = cuqi.distribution.Gaussian(np.zeros(A.domain_dim), cov=1) # x ~ N(0, 1)
    y = cuqi.distribution.Gaussian(A@x, cov=0.05**2)              # y ~ N(Ax, 0.05^2)

    # Set up Bayesian Problem
    BP = cuqi.problem.BayesianProblem(y, x).set_data(y=y_data)

    # Sample from the posterior
    samples = BP.sample_posterior(20)

    # Analyze the samples
    info.exactSolution.plot(); plt.title("Exact solution")
    y_data.plot(); plt.title("Data")
    samples.plot_mean(); plt.title("Posterior mean")
    samples.plot_std(); plt.title("Posterior standard deviation")



""" This module controls global configuration settings for the package. """
import subprocess as _subprocess

# Try detecting GPU
try:
    _subprocess.check_output('nvidia-smi')
    _detected_device = "gpu"
except:
    _detected_device = "cpu"

# Try detecting projector backend
_detected_backend = None

try:
    import astra
    _detected_backend = "astra"
except ImportError:
    pass

# Try defaulting to tigre (only if GPU is available)
if _detected_device == "gpu":
    try:
        import tigre
        _detected_backend = "tigre"
    except ImportError:
        pass

if _detected_backend is None:
    raise ImportError("No projector backend found. Please install either astra or tigre.")

PROJECTION_BACKEND = _detected_backend
""" Projection backend to use. Currently supported: "tigre", "astra". Defaults to tigre if possible, otherwise astra. """

PROJECTION_BACKEND_DEVICE = _detected_device
""" Device to use for projection backend. Currently supported: "cpu", "gpu". Defaults to GPU if present, otherwise cpu. Only relevant for astra backend. """



import cuqipy_cil
import pytest
import numpy as np
import sys

@pytest.mark.parametrize("model",
    [
        (cuqipy_cil.model.ParallelBeam2DModel()),
        (cuqipy_cil.model.FanBeam2DModel()),
        (cuqipy_cil.model.ShiftedFanBeam2DModel())
    ])
def test_model_simple(model: cuqipy_cil.model.CILModel):
    # Test that the model at least computes a forward projection.

    # Simple test input
    x = np.zeros(model.domain_geometry.fun_shape)

    # Compute forward projection
    y = model.forward(x)

    # Check that the output is the correct shape
    assert y.shape == (model.range_dim,)

    # Compute backprojection
    x2 = model.adjoint(y)

    # Check that the output is the correct shape
    assert x2.shape == (model.domain_dim,)





import cuqi
import numpy as np
import cuqipy_cil

#=============================================================================
class ParallelBeam2D(cuqi.problem.BayesianProblem):
    """ 2D parallel-beam Computed Tomography test problem using CIL.

    Parameters
    ----------
    im_size : tuple of ints
        Dimensions of image in pixels.
    
    det_count : int
        Number of detector elements.
       
    angles : ndarray
        Angles of projections, in radians.

    det_spacing : int, optional
        detector element size/spacing.
        Default is handled by the CT model.

    domain : tuple of ints, optional
        Domain of the image.
        Default is handled by the CT model.

    phantom : str or ndarray
        Phantom image to generate data from.
        If string name must match a phantom in cuqi.data.
        The string is lowercased and any hyphens are replaced 
        with underscores to match a method name in cuqi.data.

    noise_type : string
        The type of noise
        "Gaussian" - Gaussian white noise¨
        "scaledGaussian" - Scaled (by data) Gaussian noise

    noise_std : scalar
        Standard deviation of the noise

    prior : cuqi.distribution.Distribution, optional
        Distribution of the prior.
        If set posterior samples can be computed using :meth:`sample_posterior`.

    data : cuqi.array.CUQIarray, optional
        Data to be stored in testproblem.

    Attributes
    ----------
    data : CUQIarray
        Generated (noisy) data

    model : cuqipy_cil.model.CILModel
        CT Model.

    prior : cuqi.distribution.Distribution, Default None
        Distribution of the prior.

    likelihood : cuqi.likelihood.Likelihood
        Likelihood function. 
        (automatically computed from noise distribution)

    exactSolution : CUQIarray
        Exact solution (ground truth)

    exactData : CUQIarray
        Noise free data

    Methods
    ----------
    MAP()
        Compute MAP estimate of posterior.
        NB: Requires prior to be defined.

    sample_posterior(Ns)
        Sample Ns samples of the posterior.
        NB: Requires prior to be defined.


    """
    def __init__(self,
        im_size=(45, 45),
        det_count=50,
        angles=np.linspace(0, np.pi, 60),
        det_spacing=None,
        domain = None,
        phantom = "shepp-logan",
        noise_type="gaussian",
        noise_std=0.05,
        prior=None,
        data=None
        ):
        
        # CT model with default values
        model = cuqipy_cil.model.ParallelBeam2DModel(
            im_size=im_size,
            det_count=det_count,
            angles=angles,
            det_spacing=det_spacing,
            domain=domain,
        )
                      
        # Get exact phantom
        if isinstance(phantom, np.ndarray):
            if phantom.shape != model.domain_geometry.fun_shape:
                raise ValueError(f"Phantom shape does not match model domain geometry image shape {model.domain_geometry.fun_shape}.")
            x_exact = phantom
        elif isinstance(phantom, str):
            # lowercase and replace hyphens with underscores to match library method names
            phantom = phantom.lower().replace("-", "_") 
            if hasattr(cuqi.data, phantom):
                x_exact = getattr(cuqi.data, phantom)(size=model.domain_geometry.fun_shape[0])
            else:
                raise ValueError("Phantom not found in cuqi.data phantom library.")
        else:
            raise ValueError("Phantom must be a string or ndarray. See string options in cuqi.data.")
        
        x_exact = cuqi.array.CUQIarray(x_exact, is_par=False, geometry=model.domain_geometry)

        # Define prior
        if prior is None:
            prior = cuqi.distribution.Gaussian(np.zeros(model.domain_dim), cov=1, geometry = model.domain_geometry, name="x")

        # Define and add noise #TODO: Add Poisson and logpoisson
        if noise_type.lower() == "gaussian":
            data_dist = cuqi.distribution.Gaussian(model(prior), cov=noise_std**2, geometry = model.range_geometry, name="y")
        elif noise_type.lower() == "scaledgaussian":
            if data is None:
                raise ValueError("Scaled Gaussian noise requires data to be defined.")
            data_dist = cuqi.distribution.Gaussian(model(prior), cov=data*(noise_std**2), geometry = model.range_geometry, name="y")
        else:
            raise NotImplementedError("This noise type is not implemented")
        
        # Generate data
        if data is None:
            b_exact = model.forward(x_exact)
            data = data_dist(x_exact).sample()
        else:
            b_exact = None # No exact data if data is provided

        # Make likelihood
        likelihood = data_dist.to_likelihood(data)

        # Initialize CT as BayesianProblem problem
        super().__init__(likelihood, prior)

        # Store exact values
        self.exactSolution = x_exact
        self.exactData = b_exact
        self.infoString = "Noise type: Additive {} with std: {}".format(noise_type.capitalize(), noise_std)
